/* SpaceWire light driver
 * Copyright (C) 2020 All rights reserved
 *
 * Contributors:
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __SPW_LIGHT_H__
#define __SPW_LIGHT_H__

#define SPW_LIGHT_RESOURCE_NAME	"spw_light"

#include <linux/bitops.h>
#include <linux/if_vlan.h>
#include <linux/list.h>
#include <linux/netdevice.h>
#include <linux/phy.h>

#define SPW_LIGHT_SW_RESET_WATCHDOG_CNTR	10000

#define GET_BIT_VALUE(v, bit)		(((v) >> (bit)) & 0x1)

/* Control Register Bit Definitions
 */
#define SPW_CTL_RESET			BIT(0)
#define SPW_CTL_LINKSTART		BIT(2)
#define SPW_CTL_LINKAUTOSTART		BIT(3)
#define SPW_CTL_LINKDISABLE		BIT(4)
#define SPW_CTL_TICKENABLE		BIT(5)
#define SPW_CTL_IE_LINK			BIT(9)
#define SPW_CTL_IE_TIME			BIT(10)
#define SPW_CTL_IE_RXPKT		BIT(13)
#define SPW_CTL_IE_ERROR		BIT(14)

#define SPW_STS_LINK_START		BIT(0)
#define SPW_STS_LINK_CON		BIT(1)
#define SPW_STS_EDISC			BIT(2)
#define SPW_STS_EPAR			BIT(3)
#define SPW_STS_EESC			BIT(4)
#define SPW_STS_ECRED			BIT(5)
#define SPW_STS_TCODE			BIT(10)
#define SPW_STS_RXPKT			BIT(13)
#define SPW_STS_RXEMPTY			BIT(14)

#if 0
#define SPW_CTL_IE_ERROR_GET(v)		GET_BIT_VALUE(v, 14)
#endif

struct spw_light_reg {
	/* Control register
	 *  bit 0     Reset spwavl core (auto-clear)
	 *  bit 1     Reserved
	 *  bit 2     Link start
	 *  bit 3     Link autostart
	 *  bit 4     Link disable
	 *  bit 5     Enable timecode transmission through tick_in signal
	 *  bit 6:8   Reserved
	 *  bit 9     Enable interrupt on link up/down
	 *  bit 10    Enable interrupt on time code received
	 *  bit 11:12 Reserved
	 *  bit 13    Enable interrupt on RX packet
	 *  bit 14    Enable interrupt on error
	 *  bit 27:24 Reserved
	 */
	u32 control;
	/* Status register
	 *  bit 1:0   Link status: 0=off, 1=started, 2=connecting, 3=run
	 *  bit 2     Got disconnect error (sticky)
	 *  bit 3     Got parity error (sticky)
	 *  bit 4     Got escape error (sticky)
	 *  bit 5     Got credit error (sticky)
	 *  bit 6:9   Reserved
	 *  bit 10    Received timecode (sticky)
	 *  bit 11:12 Reserved
	 *  bit 13    Received packet (sticky)
	 *  bit 14    RX buffer empty after packet
	 */
	u32 status;
	/* Transmission Clock Scaler
	 *  bit 7:0   txclk division factor minus 1
	 */
	u32 clock_scaler;
	/* Timecode
	 *  bit 5:0   Last received timecode value (read-only)
	 *  bit 7:6   Control bits received with last timecode (read-only)
	 *  bit 13:8  Timecode value to send on next tick_in (auto-increment)
	 *  bit 15:14 Reserved (write as zero)
	 *  bit 16    Write '1' to send a timecode (auto-clear)
	 */
	u32 timecode;
	u32 reserved;
	u32 reserved2;
	u32 reserved3;
	/* Version
	 *  bit 0     timecode function enabled 
	 *  bit 6     rximpl (0 -> generic, 1 -> fast)
	 *  bit 7     tximpl (0 -> generic, 1 -> fast)
	 *  bit 11:8  rx FIFO size as log2
	 *  bit 15:12 tx FIFO size as log2
	 *  bit 23:16 system clock in MHz
	 *  bit 27:24 minor version (read-only)
	 *  bit 31:28 major version (read-only)
	 */
	u32 version;
};

#define spw_csroffs(a) (offsetof(struct spw_light_reg, a))

/* Transmit and Receive Command Registers Bit Definitions
 */
#define ALTERA_TSE_TX_CMD_STAT_OMIT_CRC		BIT(17)
#define ALTERA_TSE_TX_CMD_STAT_TX_SHIFT16	BIT(18)
#define ALTERA_TSE_RX_CMD_STAT_RX_SHIFT16	BIT(25)

/* Wrapper around a pointer to a socket buffer,
 * so a DMA handle can be stored along with the buffer
 */
struct tse_buffer {
	struct list_head lh;
	struct sk_buff *skb;
	dma_addr_t dma_addr;
	u32 len;
	int mapped_as_page;
};

struct altera_tse_private;

#define ALTERA_DTYPE_SGDMA 1
#define ALTERA_DTYPE_MSGDMA 2

/* standard DMA interface for SGDMA and MSGDMA */
struct altera_dmaops {
	int altera_dtype;
	int dmamask;
	void (*reset_dma)(struct altera_tse_private *);
	void (*enable_txirq)(struct altera_tse_private *);
	void (*enable_rxirq)(struct altera_tse_private *);
	void (*disable_txirq)(struct altera_tse_private *);
	void (*disable_rxirq)(struct altera_tse_private *);
	void (*clear_txirq)(struct altera_tse_private *);
	void (*clear_rxirq)(struct altera_tse_private *);
	int (*tx_buffer)(struct altera_tse_private *, struct tse_buffer *);
	u32 (*tx_completions)(struct altera_tse_private *);
	void (*add_rx_desc)(struct altera_tse_private *, struct tse_buffer *);
	u32 (*get_rx_status)(struct altera_tse_private *);
	int (*init_dma)(struct altera_tse_private *);
	void (*uninit_dma)(struct altera_tse_private *);
	void (*start_rxdma)(struct altera_tse_private *);
};

/* This structure is private to each device.
 */
struct altera_tse_private {
	struct net_device *dev;
	struct device *device;
	struct napi_struct napi;

	/* MAC address space */
	struct spw_light_reg __iomem *reg;

	/* SpW light version register */
	u32	version;

	/* mSGDMA Rx Dispatcher address space */
	void __iomem *rx_dma_csr;
	void __iomem *rx_dma_desc;
	void __iomem *rx_dma_resp;

	/* mSGDMA Tx Dispatcher address space */
	void __iomem *tx_dma_csr;
	void __iomem *tx_dma_desc;

	/* Rx buffers queue */
	struct tse_buffer *rx_ring;
	u32 rx_cons;
	u32 rx_prod;
	u32 rx_ring_size;
	u32 rx_dma_buf_sz;

	/* Tx ring buffer */
	struct tse_buffer *tx_ring;
	u32 tx_prod;
	u32 tx_cons;
	u32 tx_ring_size;

	/* Interrupts */
	u32 tx_irq;
	u32 rx_irq;
	u32 ctl_irq;

	/* RX/TX MAC FIFO configs */
	u32 tx_fifo_depth;
	u32 rx_fifo_depth;
	u32 max_mtu;

	/* Descriptor memory info for managing SGDMA */
	u32 txdescmem;
	u32 rxdescmem;
	dma_addr_t rxdescmem_busaddr;
	dma_addr_t txdescmem_busaddr;
	u32 txctrlreg;
	u32 rxctrlreg;
	dma_addr_t rxdescphys;
	dma_addr_t txdescphys;

	struct list_head txlisthd;
	struct list_head rxlisthd;

	/* MAC command_config register protection */
	spinlock_t cfg_lock;
	/* Tx path protection */
	spinlock_t tx_lock;
	/* Rx DMA & interrupt control protection */
	spinlock_t rxdma_irq_lock;

	/* PHY */
	int oldspeed;
	int oldduplex;
	int oldlink;

	/* ethtool msglvl option */
	u32 msg_enable;

	struct altera_dmaops *dmaops;
};

/* Function prototypes
 */
void altera_tse_set_ethtool_ops(struct net_device *);

static inline
u32 csrrd32(void __iomem *mac, size_t offs)
{
	void __iomem *paddr = (void __iomem *)((uintptr_t)mac + offs);
	return readl(paddr);
}

static inline
u16 csrrd16(void __iomem *mac, size_t offs)
{
	void __iomem *paddr = (void __iomem *)((uintptr_t)mac + offs);
	return readw(paddr);
}

static inline
u8 csrrd8(void __iomem *mac, size_t offs)
{
	void __iomem *paddr = (void __iomem *)((uintptr_t)mac + offs);
	return readb(paddr);
}

static inline
void csrwr32(u32 val, void __iomem *mac, size_t offs)
{
	void __iomem *paddr = (void __iomem *)((uintptr_t)mac + offs);

	writel(val, paddr);
}

static inline
void csrwr16(u16 val, void __iomem *mac, size_t offs)
{
	void __iomem *paddr = (void __iomem *)((uintptr_t)mac + offs);

	writew(val, paddr);
}

static inline
void csrwr8(u8 val, void __iomem *mac, size_t offs)
{
	void __iomem *paddr = (void __iomem *)((uintptr_t)mac + offs);

	writeb(val, paddr);
}

#endif /* __ALTERA_TSE_H__ */
